/*
 * (c) 2008-2009 Adam Lackorzynski <adam@os.inf.tu-dresden.de>
 *     economic rights: Technische Universität Dresden (Germany)
 * Copyright (C) 2017, 2018, 2019, 2020, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
/*
 * Access the I2C peripherals on the MIPS Creator CI20 board.
 */

#include <l4/devices/cpm-jz4780.h>
#include <l4/devices/gpio-jz4780.h>
#include <l4/devices/i2c-jz4780.h>
#include <l4/devices/memory.h>
#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/vbus/vbus.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>

#define REG(x) *((volatile uint32_t *) (x))

#define RTC_RTCCR   0x00
#define RTC_HCR     0x20
#define RTC_HSPR    0x34
#define RTC_WENR    0x3c
#define RTC_PWRONCR 0x3c



enum {
  PMSDA  = 30,  /* via PORTD */
  PMSCL  = 31,  /* via PORTD */
  PWM3   = 3,   /* via PORTE */
  PWM4   = 4,   /* via PORTE */
  RTCSDA = 12,  /* via PORTE */
  RTCSCL = 13,  /* via PORTE */
  DDCSCL = 24,  /* via PORTF */
  DDCSDA = 25,  /* via PORTF */
};



/* Device and resource discovery. */

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else
    return start - index;
}

static long i2c_read(void *i2c_channel, uint8_t *buf, unsigned length,
                     int stop, l4_cap_idx_t irqcap)
{
  l4_msgtag_t tag;

  jz4780_i2c_start_read(i2c_channel, buf, length, stop);

  while (!jz4780_i2c_read_done(i2c_channel))
  {
    tag = l4_irq_receive(irqcap, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4_timeout_from_us(1000)));

    if (l4_ipc_error(tag, l4_utcb()))
      return 0;

    if (jz4780_i2c_failed(i2c_channel))
      return 0;

    jz4780_i2c_read(i2c_channel);
  }

  return jz4780_i2c_have_read(i2c_channel);
}

static long i2c_write(void *i2c_channel, uint8_t *buf, unsigned length,
                      int stop, l4_cap_idx_t irqcap)
{
  l4_msgtag_t tag;
  long err;

  jz4780_i2c_start_write(i2c_channel, buf, length, stop);

  while (!jz4780_i2c_write_done(i2c_channel))
  {
    tag = l4_irq_receive(irqcap, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4_timeout_from_us(1000)));

    if ((err = l4_ipc_error(tag, l4_utcb())))
      return 0;

    if (jz4780_i2c_failed(i2c_channel))
      return 0;

    jz4780_i2c_write(i2c_channel);
  }

  return jz4780_i2c_have_written(i2c_channel);
}

static long i2c_get(void *i2c_channel, uint8_t *buffer, long length, l4_cap_idx_t irqcap)
{
  long pos;

  memset(buffer, 0, length);
  pos = i2c_read(i2c_channel, buffer, length, 1, irqcap);

  return pos;
}

static long i2c_get_reg(void *i2c_channel, uint8_t reg, uint8_t *buffer, long length, l4_cap_idx_t irqcap)
{
  buffer[0] = reg;
  i2c_write(i2c_channel, buffer, 1, 0, irqcap);

  return i2c_get(i2c_channel, buffer, length, irqcap);
}

static long i2c_set_reg(void *i2c_channel, uint8_t reg, uint8_t *buffer, long length, l4_cap_idx_t irqcap)
{
  uint8_t buf[length + 1];
  long pos;

  buf[0] = reg;
  memcpy(buf + 1, buffer, length);
  pos = i2c_write(i2c_channel, buf, length + 1, 1, irqcap);

  return pos;
}

static void i2c_report(void *i2c_channel, uint8_t *buffer, long length)
{
  if (!jz4780_i2c_failed(i2c_channel))
  {
    printf("Read: ");
    while (length--)
      printf("%02x", *(buffer++));
    printf("\n");
  }
  else
    printf("Read failed.\n");
}

static void i2c_scan(void *i2c_channel, l4_cap_idx_t irqcap)
{
  uint8_t buf[1];
  unsigned int address;

  for (address = 0; address < 0x20; address++)
    printf("%02x ", address);
  printf("\n");

  for (address = 0; address < 0x20; address++)
    printf("-- ");

  for (address = 0; address < 0x80; address++)
  {
    if ((address % 32) == 0)
      printf("\n");

    jz4780_i2c_set_target(i2c_channel, address);
    i2c_get(i2c_channel, buf, 1, irqcap);

    if (!jz4780_i2c_failed(i2c_channel))
      printf("%02x ", address);
    else
      printf("?? ");
  }

  printf("\n");
  for (address = 0; address < 0x20; address++)
    printf("-- ");
  printf("\n\n");
}

static void pmic_dump(void *i2c_channel, l4_cap_idx_t irqcap)
{
  uint8_t regs[] = {0x00, 0x01, 0x10, 0x12, 0x20, 0x22, 0x30, 0x32,
                    0x40, 0x41, 0x50, 0x51, 0x60, 0x61, 0x70, 0x71,
                    0x80, 0x81, 0x91};
  uint8_t buf[1];
  long pos;
  unsigned i;

  for (i = 0; i < sizeof(regs); i++)
  {
    printf("PMIC %02x:\n", regs[i]);
    jz4780_i2c_set_target(i2c_channel, 0x5a);
    pos = i2c_get_reg(i2c_channel, regs[i], buf, 1, irqcap);
    i2c_report(i2c_channel, buf, pos);
  }
}

static void rtc_dump(void *i2c_channel, l4_cap_idx_t irqcap, uint8_t *rtcregs)
{
  /* Attempt to read from address 0x51 for RTC. */

  jz4780_i2c_set_target(i2c_channel, 0x51);
  i2c_get_reg(i2c_channel, 0x00, rtcregs, 16, irqcap);
  i2c_report(i2c_channel, rtcregs, 16);
}

static void rtc_datetime(uint8_t *rtcregs)
{
  char *weekdays[] = {"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"};
  printf("%s %d%d-%d%d-%d%d %d%d:%d%d:%d%d\n",
    weekdays[rtcregs[0x06] & 0x07],
    (rtcregs[0x08] & 0xf0) >> 4,
    rtcregs[0x08] & 0x0f,
    (rtcregs[0x07] & 0x10) >> 4,
    rtcregs[0x07] & 0x0f,
    (rtcregs[0x05] & 0x30) >> 4,
    rtcregs[0x05] & 0x0f,
    (rtcregs[0x04] & 0x30) >> 4,
    rtcregs[0x04] & 0x0f,
    (rtcregs[0x03] & 0x70) >> 4,
    rtcregs[0x03] & 0x0f,
    (rtcregs[0x02] & 0x70) >> 4,
    rtcregs[0x02] & 0x0f);
}

static int rtc_update(void *i2c_channel, l4_cap_idx_t irqcap, uint8_t *rtcregs)
{
  jz4780_i2c_set_target(i2c_channel, 0x51);
  return i2c_set_reg(i2c_channel, 0x02, rtcregs + 2, 7, irqcap);
}

static void rtc_write_ready(l4_addr_t rtc_base)
{
  while (!(REG(rtc_base + RTC_RTCCR) & 0x80));
}

static void rtc_write_enable(l4_addr_t rtc_base)
{
  rtc_write_ready(rtc_base);
  REG(rtc_base + RTC_WENR) = 0xa55a;
  while (!(REG(rtc_base + RTC_WENR) & 0x80000000));
}

int main(void)
{
  long err;

  /* Peripheral memory. */

  l4_addr_t gpio_base = 0, gpio_base_end = 0;
  l4_addr_t i2c_base = 0, i2c_base_end = 0;
  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  l4_addr_t rtc_base = 0, rtc_base_end = 0;
  l4_addr_t port_d, port_d_end;
  l4_addr_t port_e, port_e_end;

  /* Peripheral abstractions. */

  void *gpio_port_d, *gpio_port_e;
  void *i2c, *cpm;
  void *i2c0, *i2c4;

  /* IRQ resources. */

  l4_uint32_t i2c_irq_start = 0, i2c_irq_end = 0;
  l4_cap_idx_t icucap, irq0cap, irq4cap;

  /* RTC registers. */

  uint8_t rtcregs[16];
  int i;

  /* Obtain capabilities for the interrupt controller and an interrupt. */

  irq0cap = l4re_util_cap_alloc();
  irq4cap = l4re_util_cap_alloc();
  icucap = l4re_env_get_cap("icu");

  if (l4_is_invalid_cap(icucap))
  {
    printf("No 'icu' capability available in the virtual bus.\n");
    return 1;
  }

  if (l4_is_invalid_cap(irq0cap) || l4_is_invalid_cap(irq4cap))
  {
    printf("Capabilities not available for interrupts.\n");
    return 1;
  }

  /* Obtain resource details describing the interrupt for I2C channel 4. */

  printf("Access IRQ...\n");

  if (get_irq("jz4780-i2c", &i2c_irq_start, &i2c_irq_end) < 0)
    return 1;

  printf("IRQ range at %d...%d.\n", i2c_irq_start, i2c_irq_end);

  /* Obtain resource details describing I/O memory. */

  printf("Access GPIO...\n");

  if (get_memory("jz4780-gpio", &gpio_base, &gpio_base_end) < 0)
    return 1;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);

  printf("Access CPM...\n");

  if (get_memory("jz4780-cpm", &cpm_base, &cpm_base_end) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  printf("Access I2C...\n");

  if (get_memory("jz4780-i2c", &i2c_base, &i2c_base_end) < 0)
    return 1;

  printf("I2C at 0x%lx...0x%lx.\n", i2c_base, i2c_base_end);

  printf("Access RTC...\n");

  if (get_memory("jz4780-rtc", &rtc_base, &rtc_base_end) < 0)
    return 1;

  printf("RTC at 0x%lx...0x%lx.\n", rtc_base, rtc_base_end);

  /* Create interrupt objects. */

  err = l4_error(l4_factory_create_irq(l4re_global_env->factory, irq0cap)) ||
        l4_error(l4_factory_create_irq(l4re_global_env->factory, irq4cap));

  if (err)
  {
    printf("Could not create IRQ object: %lx\n", err);
    return 1;
  }

  /* Bind interrupt objects to IRQ numbers. */

  err = l4_error(l4_icu_bind(icucap,
                             item_in_range(i2c_irq_start, i2c_irq_end, 0),
                             irq0cap)) ||
        l4_error(l4_icu_bind(icucap,
                             item_in_range(i2c_irq_start, i2c_irq_end, 4),
                             irq4cap));

  if (err)
  {
    printf("Could not bind IRQ to the ICU: %ld\n", err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  err = l4_error(l4_rcv_ep_bind_thread(irq0cap, l4re_env()->main_thread, 0)) ||
        l4_error(l4_rcv_ep_bind_thread(irq4cap, l4re_env()->main_thread, 4));

  if (err)
  {
    printf("Could not attach to IRQs: %ld\n", err);
    return 1;
  }

  /* Configure pins. */

  port_d = gpio_base + 0x300;
  port_d_end = port_d + 0x100;
  port_e = gpio_base + 0x400;
  port_e_end = port_e + 0x100;

  printf("PORTD at 0x%lx...0x%lx.\n", port_d, port_d_end);
  printf("PORTE at 0x%lx...0x%lx.\n", port_e, port_e_end);

  gpio_port_d = jz4780_gpio_init(port_d, port_d_end, 32, 0xffff4fff, 0x0000b000);
  gpio_port_e = jz4780_gpio_init(port_e, port_e_end, 32, 0xfffff37c, 0x00000483);

  printf("Set up GPIO pins...\n");

  jz4780_gpio_config_pad(gpio_port_d, PMSCL, Function_alt, 0);
  jz4780_gpio_config_pad(gpio_port_d, PMSDA, Function_alt, 0);

  jz4780_gpio_config_pad(gpio_port_e, RTCSCL, Function_alt, 1);
  jz4780_gpio_config_pad(gpio_port_e, RTCSDA, Function_alt, 1);

  /* Obtain CPM and I2C objects. */

  cpm = jz4780_cpm_init(cpm_base);

  /* Attempt to set the PCLK source to SCLK_A. */

  jz4780_cpm_set_source_clock(cpm, Clock_pclock, Clock_main);
  printf("Peripheral clock: %lld\n", jz4780_cpm_get_frequency(cpm, Clock_pclock));

  /* Obtain I2C reference. */

  i2c = jz4780_i2c_init(i2c_base, i2c_base_end, cpm, 100000); /* 100 kHz */
  i2c0 = jz4780_i2c_get_channel(i2c, 0);
  i2c4 = jz4780_i2c_get_channel(i2c, 4);

  printf("Scan I2C0...\n");
  i2c_scan(i2c0, irq0cap);

  printf("Scan I2C4...\n");
  i2c_scan(i2c4, irq4cap);

  printf("PMIC...\n");
  pmic_dump(i2c0, irq0cap);

  printf("RTC...\n");
  rtc_dump(i2c4, irq4cap, rtcregs);

  rtcregs[0x02] = 0x56;
  rtcregs[0x03] = 0x34;
  rtcregs[0x04] = 0x12;
  rtcregs[0x05] = 0x25; /* 25th */
  rtcregs[0x06] = 0x06; /* Saturday */
  rtcregs[0x07] = 0x01; /* January */
  rtcregs[0x08] = 0x20; /* 2020 */

  printf("Updated %d registers.\n", rtc_update(i2c4, irq4cap, rtcregs));

  for (i = 0; i < 3; i++)
  {
    rtc_dump(i2c4, irq4cap, rtcregs);
    rtc_datetime(rtcregs);
    sleep(1);
  }

  /* Investigate the internal RTC registers. */

  printf("Control: %08x\n", REG(rtc_base + RTC_RTCCR));
  printf("Hibernate: %08x\n", REG(rtc_base + RTC_HCR));
  printf("Scratchpad: %08x\n", REG(rtc_base + RTC_HSPR));
  printf("Power on control: %08x\n", REG(rtc_base + RTC_PWRONCR));
  printf("Write enable: %08x\n", REG(rtc_base + RTC_WENR));

  rtc_write_enable(rtc_base);
  rtc_write_ready(rtc_base);
  REG(rtc_base + RTC_HSPR) = 0x12345678;

  for (i = 0; i < 3; i++)
  {
    printf("Scratchpad: %08x\n", REG(rtc_base + RTC_HSPR));
    sleep(1);
  }

  //jz4780_i2c_disable(i2c0);
  //jz4780_i2c_disable(i2c4);

  /* Detach from the interrupt. */

  err = l4_error(l4_irq_detach(irq0cap)) || l4_error(l4_irq_detach(irq4cap));

  if (err)
    printf("Error detaching from IRQ objects: %ld\n", err);

  printf("Done.\n");

  return 0;
}
